REM
REM  KittyPOP Copyright (C) Mateusz Viste
REM
REM  This program is free software: you can redistribute it and/or modify
REM  it under the terms of the GNU General Public License as published by
REM  the Free Software Foundation, either version 3 of the License, or
REM  (at your option) any later version.
REM
REM  This program is distributed in the hope that it will be useful,
REM  but WITHOUT ANY WARRANTY; without even the implied warranty of
REM  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
REM  GNU General Public License for more details.
REM
REM  You should have received a copy of the GNU General Public License
REM  along with this program.  If not, see <http://www.gnu.org/licenses/>.
REM

REM Include all required libs
#INCLUDE ONCE "vbcompat.bi"
#INCLUDE ONCE "crt/linux/unistd.bi" ' Required for GetPID()
#INCLUDE ONCE "md5.bi"

REM Declare all constants
CONST CRLF AS STRING = CHR(13,10)
CONST pVer AS STRING = "0.50"
CONST pDate AS STRING = "2009"

REM Declare all shared variables
REDIM SHARED Maillist(1 TO 2) AS STRING
REDIM SHARED Mailsize(1 TO 2) AS INTEGER
REDIM SHARED UserNameList(1 TO 2) AS STRING
REDIM SHARED UserPassList(1 TO 2) AS STRING
DIM SHARED AS INTEGER Mailnumber, FakeBoxTrap, AuthTries = 0, UserListSize
DIM SHARED AS STRING Maildir, Spooldir, LogFile, APOPbanner
DIM SHARED AS BYTE AuthState, LogLevel, KeepMails
DIM SHARED AS STRING Username, UserPass


REM Below are all functions and subroutines


FUNCTION GetToken(Source AS STRING, index AS INTEGER, Delimiter AS STRING = " ") AS STRING
    DIM AS INTEGER cur, lend = Len(delimiter), last = 1 - lend, cnt = 0
    DO
        cur = INSTR(last + lend, source, delimiter)
        IF cur > 0 THEN
            cnt += 1
            IF cnt = index THEN RETURN MID(source, last + lend, cur - last - lend)
        ELSE
            IF index = cnt + 1 THEN RETURN RIGHT(source, Len(source) - last)
        END IF
        last = cur
    LOOP UNTIL cur = 0
    RETURN ""
END FUNCTION


FUNCTION MD5Sum(RawData AS STRING) AS STRING
  DIM AS STRING Result
  Result = MD5_FromString(RawData)
  RETURN Result
END FUNCTION


FUNCTION CheckTrimBOM(WorkString AS STRING) AS STRING  ' Remove any BOM UTF8 header from the line
  DIM Result AS STRING = WorkString
  DIM WorkStringLen AS INTEGER = LEN(WorkString)
  IF WorkStringLen >= 3 THEN
    IF MID(WorkString, 1, 3) = CHR(239) + CHR(187) + CHR(191) THEN
      IF WorkStringLen > 3 THEN
          Result = MID(WorkString, 4)
        ELSE
          Result = ""
      END IF
    END IF
  END IF
  RETURN Result
END FUNCTION


SUB LogLine(Logdata AS STRING)
  DIM AS INTEGER FileHandler
  DIM ShortMonth(1 TO 12) AS STRING*3 => {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"}
  DIM AS DOUBLE DateStamp
  DIM AS STRING LogLineBuf
  IF LogLevel > 0 THEN
    DateStamp = NOW
    LogLineBuf = "[" & GetPID() & "] " + FORMAT(DAY(DateStamp), "00") + " " + ShortMonth(MONTH(DateStamp)) + " " + FORMAT(YEAR(DateStamp), "0000") + " " + FORMAT(HOUR(DateStamp), "00") + ":" + FORMAT(MINUTE(DateStamp), "00") + ":" + FORMAT(SECOND(DateStamp), "00") + " > " + LogData
    FileHandler = FREEFILE
    OPEN LogFile FOR APPEND AS FileHandler
    PRINT #FileHandler, LogLineBuf
    CLOSE #FileHandler
  END IF
END SUB


SUB RunCAPA()
  PRINT "+OK Capability list follows:" + CRLF;
  IF AuthState < 2 THEN
      PRINT "USER" + CRLF;
    ELSE
      PRINT "TOP" + CRLF;
  END IF
  PRINT "AUTH-RESP-CODE" + CRLF;
  PRINT "IMPLEMENTATION KittyPOP v" + pVer + CRLF;
  PRINT "." + CRLF;
END SUB


SUB CheckInput(BYREF ClientData AS STRING)
  DIM AS INTEGER x
  IF LEN(ClientData) > 1024 THEN
    LogLine("Overflow attempt detected! Client's request is " & LEN(ClientData) & " bytes long. Connection closed.")
    END
  END IF
  FOR x = 0 TO LEN(ClientData) - 1
    IF ClientData[x] > 127 THEN
      LogLine("Illegal (hi-ASCII) character detected in the request! Connection closed.")
    END IF
    IF ClientData[x] = 0 THEN
      LogLine("Nul character detected in the request! Connection closed.")
    END IF
  NEXT x
END SUB


SUB FakeMailbox()
  DIM AS STRING LineBuff, PopCommand, PopParam
  AuthState = 2
  LogLine("The brute-forcer has been redirected into the fake mailbox.")
  PRINT "+OK Authentication OK." + CRLF;
  DO
    LINE INPUT #1, LineBuff
    CheckInput(LineBuff)
    SLEEP 250, 1
    PopCommand = GetToken(LineBuff, 1)
    SELECT CASE UCASE(PopCommand)
      CASE "STAT"
        PRINT "+OK 0 0" + CRLF;
      CASE "TOP"
        PRINT "-ERR No such message." + CRLF;
      CASE "LIST"
        PRINT "+OK 0 messages." + CRLF;
        PRINT "." + CRLF;
      CASE "RSET"
        PRINT "+OK State reset done." + CRLF;
      CASE "UIDL"
        PRINT "-ERR No such message." + CRLF;
      CASE "RETR"
        PRINT "-ERR No such message." + CRLF;
      CASE "DELE"
        PRINT "-ERR No such message." + CRLF;
      CASE "QUIT"
        PRINT "+OK See you later!" + CRLF;
        CLOSE #1
        END
      CASE "NOOP"
        PRINT "+OK Yup, still here." + CRLF;
      CASE "CAPA"
        RunCAPA()
      CASE "LAST"
        PRINT "-ERR The LAST command is obsolete. Go get a modern POP3 client!" + CRLF;
      CASE ELSE
        PRINT "-ERR Unrecognised command!" + CRLF;
    END SELECT
  LOOP
END SUB


FUNCTION DoesMsgExist(MsgNumber AS INTEGER) AS BYTE
  REM  Returns 1 if the given message exists, -1 if it has been deleted, and 0 if it
  REM  doesn't existed at all (in the current session).
  DIM AS INTEGER x
  IF MsgNumber < 1 OR MsgNumber > Mailnumber THEN
      RETURN 0
    ELSE
      IF Mailsize(MsgNumber) < 0 THEN
          RETURN 0
        ELSE
          RETURN 1
      END IF
  END IF
END FUNCTION


FUNCTION CountMails() AS INTEGER
  REM  Returns the total amount of (non-deleted) mails.
  DIM AS INTEGER x, Result = 0
  FOR x = 1 TO Mailnumber
    IF Mailsize(x) >= 0 THEN
      Result += 1
    END IF
  NEXT x
  RETURN Result
END FUNCTION


FUNCTION CheckAuth(Userlogin AS STRING, Userpassword AS STRING, AuthMethod AS STRING = "PASS") AS BYTE
  REM Returns 1 if given credentials are okay, 0 otherwise.
  DIM Result AS BYTE = 0
  DIM AS STRING ExpectedPassword
  DIM AS INTEGER x
  FOR x = 1 TO UserListSize
    IF UserNameList(x) = Userlogin THEN ExpectedPassword = UserPassList(x)
  NEXT x
  SELECT CASE AuthMethod
    CASE "PASS"
      IF Userpassword = ExpectedPassword THEN Result = 1
    CASE "APOP"
      IF Userpassword = MD5sum(APOPbanner + ExpectedPassword) THEN Result = 1
  END SELECT
  IF Result = 1 THEN
  WHILE INSTR(UCASE(Maildir), "$USER") > 0
    Maildir = MID(Maildir, 1, INSTR(UCASE(Maildir), "$USER") - 1) + Userlogin + MID(Maildir, INSTR(UCASE(Maildir), "$USER") + 5)
  WEND
  END IF
  RETURN Result
END FUNCTION


FUNCTION GetFileSize(FileToCheck AS STRING) AS INTEGER
  DIM AS UINTEGER FileSize
  DIM AS INTEGER FileHook
  FileHook = FREEFILE
  OPEN FileToCheck FOR BINARY AS FileHook
  FileSize = LOF(FileHook)
  CLOSE #FileHook
  RETURN FileSize
END FUNCTION


SUB LoadMaillist()
  DIM AS STRING NextFile
  Mailnumber = 0
  NextFile = DIR(Maildir + "new/*", &h23)
  WHILE LEN(NextFile) > 0
    IF LEFT(NextFile, 1) <> "." THEN  ' ignore any maildir file begining by a dot
        Mailnumber += 1
        Maillist(Mailnumber) = NextFile
        Mailsize(Mailnumber) = GetFileSize(Maildir + "new/" + NextFile)
        IF Mailsize(Mailnumber) = 0 THEN
          LogLine("ERROR: Message " + Maildir + "new/" + NextFile + " has a nul size! This message will be ignored.")
          Mailnumber -= 1
        END IF
        IF UBOUND(Maillist) = Mailnumber THEN
          REDIM PRESERVE Maillist(1 TO Mailnumber * 2) AS STRING
          REDIM PRESERVE Mailsize(1 TO Mailnumber * 2) AS INTEGER
        END IF
      ELSE
        LogLine("Warning: Hidden message found: " + Maildir + "new/" + NextFile + "! This message will be ignored.")
    END IF
    NextFile = DIR()
  WEND
END SUB


FUNCTION LoadUserlist() AS BYTE
  DIM AS BYTE Result = 0
  DIM AS INTEGER FileHook
  DIM AS STRING LineBuff, AuthFile
  AuthFile = "/etc/kittypop.auth"
  FileHook = FREEFILE
  IF OPEN(AuthFile, FOR INPUT, AS #FileHook) = 0 THEN
      WHILE NOT EOF(FileHook)
        LINE INPUT #FileHook, LineBuff
        IF LEN(LineBuff) > 0 THEN
          IF LEFT(LineBuff, 1) <> "#" THEN
            IF UserListSize = UBOUND(UserNameList) THEN
              REDIM PRESERVE UserNameList(1 TO UserListSize * 2) AS STRING
              REDIM PRESERVE UserPassList(1 TO UserListSize * 2) AS STRING
            END IF
            UserListSize += 1
            UserNameList(UserListSize) = GetToken(LineBuff, 1)
            UserPassList(UserListSize) = GetToken(LineBuff, 2)
            IF LEN(UserNameList(UserListSize)) = 0 OR LEN(UserPassList(UserListSize)) = 0 THEN
              LogLine("An invalid entry has been found in the " + AuthFile + " file!")
              Result = -1
            END IF
          END IF
        END IF
      WEND
      CLOSE #FileHook
    ELSE
      LogLine("ERROR! Loading " + AuthFile + " file failed!")
      Result = -1
  END IF
  IF Result = 0 AND UserListSize = 0 THEN
    LogLine("ERROR! No users found in the " + AuthFile + " file!")
    Result = -1
  END IF
  RETURN Result
END FUNCTION


SUB LockAccount()
  DIM AS INTEGER FileHook
  DIM AS BYTE Result
  FileHook = FREEFILE
  IF OPEN(Maildir + "lockfile", FOR INPUT, AS FileHook) = 0 THEN
      CLOSE #FileHook
      LogLine("Connection closed: mailbox already LOCKed.")
      PRINT "-ERR [IN-USE] Unable to obtain a lock. You can't use several simulteanous sessions!" + CRLF;
      CLOSE #1
      END
    ELSE
      IF OPEN(Maildir + "lockfile", FOR OUTPUT, AS FileHook) = 0 THEN
          PRINT #FileHook, GetPID()
          CLOSE #FileHook
        ELSE
          LogLine("ERROR: Couldn't write a LOCK! Connection closed.")
          PRINT "-ERR [SYS/PERM] Unable to write a lock." + CRLF;
          CLOSE #1
          END
      END IF
  END IF
END SUB


FUNCTION UnlockAccount() AS BYTE
   REM Returns 0 if unlock succeeded, non-zero otherwise.
   DIM AS INTEGER Result
   Result = KILL(Maildir + "lockfile")  ' Remove the lock file
   RETURN Result
END FUNCTION


SUB LoadConfig()
  DIM AS STRING CfgReturnString, CfgTmpBuffer, CfgToken, CfgValue, CfgFile
  DIM AS INTEGER CfgFileHandler, CfgColonPos, Counter
  CfgFile = "/etc/kittypop.conf"
  CfgFileHandler = FREEFILE
  CfgReturnString = ""
  Counter = 0
  REM Load default values...
  LogLevel = 1
  KeepMails = 0
  Maildir = "/var/mail/$USER/"
  FakeBoxTrap = 13
  LogFile = "/var/log/kittypop.log"
  REM Check the config file...
  IF OPEN(CFGfile, FOR INPUT, AS #CfgFileHandler) = 0 THEN
      DO
        Counter += 1
        LINE INPUT #CfgFileHandler, CfgTmpBuffer
        CfgTmpBuffer = CheckTrimBOM(CfgTmpBuffer)
        IF MID(TRIM(CfgTmpBuffer), 1, 1) <> "#" THEN
          CfgColonPos = INSTR(CfgTmpBuffer, "=")
          CfgToken = TRIM(MID(CfgTmpBuffer, 1, CfgColonPos - 1))
          CfgValue = TRIM(MID(CfgTmpBuffer, CfgColonPos + 1))
          IF LEN(CfgToken) > 0 AND LEN(CfgValue) > 0 THEN
            SELECT CASE UCASE(CfgToken)
              CASE "LOGLEVEL"
                LogLevel = VAL(CfgValue)
              CASE "FAKEBOXTRAP"
                FakeBoxTrap = VAL(CfgValue)
              CASE "LOGFILE"
                LogFile = CfgValue
              CASE "MAILDIR"
                Maildir = CfgValue
                IF RIGHT(Maildir, 1) <> "/" THEN Maildir += "/"
              CASE "KEEPMAILS"
                IF TRIM(CfgValue) = "1" THEN KeepMails = 1 ELSE KeepMails = 0
            END SELECT
          END IF
        END IF
      LOOP UNTIL EOF(CfgFileHandler) OR Counter = 255
      CLOSE #CfgFileHandler
    ELSE
      LogLine("ERROR: Loading the " + CfgFile + " configuration file failed! Fallbacking to default values...")
  END IF
END SUB


REM    * * *  The real program starts here!  * * *


DIM AS STRING PopCommand, LineBuff
DIM AS INTEGER x, y, z
DIM PopArg(1 TO 5) AS STRING

LoadConfig()
IF LEN(ENVIRON("REMOTE_ADDR")) = 0 THEN
    LogLine("Connection received from " + ENVIRON("REMOTE_HOST"))
  ELSE
    LogLine("Connection received from " + ENVIRON("REMOTE_ADDR"))
END IF

SLEEP 500, 1    ' It's often a good idea to hold the remote client for a moment...

IF LoadUserlist() <> 0 THEN
  PRINT "-ERR [SYS/PERM] INTERNAL SERVER ERROR. Please come back later. Sorry." + CRLF;
  END
END IF

OPEN CONS FOR INPUT AS #1

APOPbanner = "<" & GetPID & "." & NOW & "@KittyPOP>"

PRINT "+OK Welcome to this KittyPOP server! " + APOPbanner + CRLF;

WHILE NOT AuthState = 2
  LINE INPUT #1, LineBuff
  CheckInput(LineBuff)
  SLEEP 250, 1
  SELECT CASE AuthState
    CASE 0
      SELECT CASE UCASE(GetToken(LineBuff, 1))
        CASE "USER"
          IF LEN(GetToken(LineBuff, 2)) = 0 THEN
              PRINT "-ERR Missing argument!" + CRLF;
            ELSE
              IF LEN(GetToken(LineBuff, 3)) > 0 THEN
                  PRINT "-ERR Too many arguments!" + CRLF;
                ELSE
                  AuthState = 1
                  Username = GetToken(LineBuff, 2)
                  PRINT "+OK Tell me more." + CRLF;
              END IF
          END IF
        CASE "QUIT"
          PRINT "+OK See you later!" + CRLF;
          LogLine("Quit command received")
          CLOSE #1
          END
        CASE "CAPA"
          RunCAPA()
        CASE "APOP"
          IF LEN(GetToken(LineBuff, 3)) = 0 THEN
              PRINT "-ERR Missing argument!" + CRLF;
            ELSE
              IF LEN(GetToken(LineBuff, 4)) > 0 THEN
                  PRINT "-ERR Too many arguments!" + CRLF;
                ELSE
                  Username = GetToken(LineBuff, 2)
                  Userpass = GetToken(LineBuff, 3)
                  IF CheckAuth(Username, Userpass, "APOP") = 1 THEN
                      LockAccount()
                      AuthState = 2
                      LogLine("Successfull APOP authentication for user " + Username)
                      PRINT "+OK Authentication passed." + CRLF;
                    ELSE
                      AuthTries += 1
                      IF AuthTries = FakeBoxTrap THEN FakeMailbox()
                      PRINT "-ERR [AUTH] Authentication failed. Check carefully your login and password..." + CRLF;
                      LogLine("APOP auth failed for user " + Username + ".")
                  END IF
              END IF
          END IF
        CASE "LAST"
          PRINT "-ERR The LAST command is obsolete. Go get a modern POP3 client!" + CRLF;
        CASE ELSE
          PRINT "-ERR Nope, that doesn't work." + CRLF;
      END SELECT
    CASE 1
      SELECT CASE UCASE(GetToken(LineBuff, 1))
        CASE "PASS"
          IF LEN(GetToken(LineBuff, 3)) > 0 THEN
              PRINT "-ERR Too many arguments!" + CRLF;
            ELSE
              IF LEN(GetToken(LineBuff, 2)) = 0 THEN
                  PRINT "-ERR Missing argument!" + CRLF;
                ELSE
                  Userpass = GetToken(LineBuff, 2)
                  IF CheckAuth(Username, Userpass) = 1 THEN
                      AuthState = 2
                      LogLine("Successfull (clear-text) authentication for user " + Username)
                      LockAccount()
                      PRINT "+OK Authentication passed." + CRLF;
                    ELSE
                      AuthTries += 1
                      AuthState = 0
                      IF AuthTries = FakeBoxTrap THEN FakeMailbox()
                      PRINT "-ERR [AUTH] Authentication failed. Check carefully your login and password..." + CRLF;
                      LogLine("Clear-text auth failed for user " + Username + ".")
                  END IF
              END IF
          END IF
        CASE "QUIT"
          PRINT "+OK See you later!" + CRLF;
          LogLine("Quit command received")
          CLOSE #1
          END
        CASE "CAPA"
          RunCAPA()
        CASE "LAST"
          PRINT "-ERR The LAST command is obsolete. Go get a modern POP3 client!" + CRLF;
        CASE ELSE
          PRINT "-ERR Nope, that doesn't work. What about a PASS now?" + CRLF;
      END SELECT
  END SELECT
WEND

REM   *  From now on, the user is authenticated  *

LoadMaillist()

DO
  LINE INPUT #1, LineBuff
  CheckInput(LineBuff)
  PopCommand = GetToken(LineBuff, 1)
  FOR x = 1 TO 5
    PopArg(x) = GetToken(LineBuff, x + 1)
  NEXT x
  SELECT CASE UCASE(PopCommand)
    CASE "STAT"
      y = 0
      z = 0
      FOR x = 1 TO Mailnumber
        IF DoesMsgExist(x) = 1 THEN
          y += Mailsize(x)
          z += 1
        END IF
      NEXT x
      PRINT "+OK " & z & " " & y & CRLF;
    CASE "TOP"
      IF LEN(PopArg(2)) = 0 THEN
          PRINT "-ERR Too few arguments!" + CRLF;
        ELSE
          IF DoesMsgExist(VAL(PopArg(1))) <> 1 THEN
              PRINT "-ERR No such message." + CRLF;
            ELSE
              IF VAL(PopArg(2)) < 0 THEN
                  PRINT "-ERR Invalid argument!" + CRLF;
                ELSE
                  IF LogLevel > 1 THEN LogLine("Retrieving " & VAL(PopArg(2)) & " body lines of message #" & VAL(PopArg(1)) & ".")
                  PRINT "+OK Here is your stuff:" + CRLF;
                  y = FREEFILE
                  x = 0
                  IF OPEN(Maildir + "new/" + Maillist(VAL(PopArg(1))), FOR INPUT, AS #y) = 0 THEN
                      WHILE x <= VAL(PopArg(2)) AND NOT EOF(y)
                        LINE INPUT #y, LineBuff
                        IF x > 0 THEN x += 1
                        IF LEN(LineBuff) = 0 AND x = 0 THEN x += 1
                        IF LEFT(LineBuff, 1) = "." THEN ' dot-stuff any line beginning by a dot.
                            PRINT "." + LineBuff + CRLF;
                          ELSE
                            PRINT LineBuff + CRLF;
                        END IF
                      WEND
                      CLOSE #y
                      PRINT "." + CRLF;
                    ELSE
                      PRINT "-ERR [SYS/PERM] Unable to open the requested message!" + CRLF;
                      LogLine("ERROR: Unable to open the " + Maildir + "new/" + Maillist(VAL(PopArg(1))) + " file!")
                  END IF
              END IF
          END IF
      END IF
    CASE "LIST"
      IF LEN(PopArg(2)) > 0 THEN
          PRINT "-ERR Too many arguments!" + CRLF;
        ELSE
          IF LEN(PopArg(1)) = 0 THEN
              PRINT "+OK " & CountMails & " message(s)." + CRLF;
              FOR x = 1 TO Mailnumber
                IF DoesMsgExist(x) = 1 THEN
                  PRINT "" & x & " " & Mailsize(x) & CRLF;
                END IF
              NEXT x
              PRINT "." + CRLF;
            ELSE
              IF DoesMsgExist(VAL(PopArg(1))) <> 1 THEN
                  PRINT "-ERR No such message!" + CRLF;
                ELSE
                  PRINT "+OK " & VAL(PopArg(1)) & " " & MailSize(VAL(PopArg(1))) & CRLF;
              END IF
          END IF
      END IF
    CASE "RSET"
      FOR x = 1 TO Mailnumber
        IF Mailsize(x) < 0 THEN Mailsize(x) = ABS(Mailsize(x))
      NEXT x
      IF LogLevel > 1 THEN LogLine("Reseting mailbox' state.")
      PRINT "+OK " & CountMails() & " messages in the queue." + CRLF;
    CASE "UIDL"
      IF LEN(PopArg(2)) > 0 THEN
          PRINT "-ERR Too many arguments!" + CRLF;
        ELSE
          IF LEN(PopArg(1)) > 0 THEN
              IF DoesMsgExist(VAL(PopArg(1))) = 1 THEN
                  PRINT "+OK " & VAL(PopArg(1)) & " " & Maillist(VAL(PopArg(1))) + CRLF;
                ELSE
                  PRINT "-ERR No such message!" + CRLF;
              END IF
            ELSE
              PRINT "+OK UIDL map follows..." + CRLF;
              FOR x = 1 TO Mailnumber
                IF DoesMsgExist(x) = 1 THEN PRINT "" & x & " " & Maillist(x) + CRLF;
              NEXT x
              PRINT "." + CRLF;
          END IF
      END IF
    CASE "RETR"
      IF LEN(PopArg(2)) > 0 THEN
          PRINT "-ERR Too many arguments!" + CRLF;
        ELSE
          IF LEN(PopArg(1)) = 0 THEN
              PRINT "-ERR Missing argument!" + CRLF;
            ELSE
              IF DoesMsgExist(VAL(PopArg(1))) <> 1 THEN
                  PRINT "-ERR No such message!" + CRLF;
                ELSE
                  IF LogLevel > 1 THEN LogLine("Retrieving message #" & VAL(PopArg(1)) & ".")
                  PRINT "+OK Here comes your message..." + CRLF;
                  y = FREEFILE
                  IF OPEN(Maildir + "new/" + Maillist(VAL(PopArg(1))), FOR INPUT, AS #y) = 0 THEN
                      WHILE NOT EOF(y)
                        LINE INPUT #y, LineBuff
                        IF LEFT(LineBuff, 1) = "." THEN ' dot-stuff any line beginning by a dot.
                            PRINT "." + LineBuff + CRLF;
                          ELSE
                            PRINT LineBuff + CRLF;
                        END IF
                      WEND
                      CLOSE #y
                      PRINT "." + CRLF;
                    ELSE
                      PRINT "-ERR [SYS/PERM] Unable to open the requested message!" + CRLF;
                      LogLine("ERROR: Unable to open the " + Maildir + "new/" + Maillist(VAL(PopArg(1))) + " file!")
                  END IF
              END IF
          END IF
      END IF
    CASE "DELE"
      IF LEN(PopArg(2)) > 0 THEN
          PRINT "-ERR Too many arguments!" + CRLF;
        ELSE
          IF LEN(PopArg(1)) = 0 THEN
              PRINT "-ERR Missing argument!" + CRLF;
            ELSE
              IF DoesMsgExist(VAL(PopArg(1))) <> 1 THEN
                  PRINT "-ERR No such message!" + CRLF;
                ELSE
                  IF LogLevel > 1 THEN LogLine("Marking message #" & VAL(PopArg(1)) & " as deleted.")
                  Mailsize(VAL(PopArg(1))) = 0 - Mailsize(VAL(PopArg(1)))
                  PRINT "+OK Message has been marked as deleted." + CRLF;
              END IF
          END IF
      END IF
    CASE "QUIT"
      LogLine("Quit command received.")
      y = 0
      FOR x = 1 TO Mailnumber
        IF Mailsize(x) < 0 THEN
          y += 1
          IF KeepMails = 1 THEN
              z = NAME(Maildir + "new/" + Maillist(x), Maildir + "cur/" + Maillist(x))
              IF z <> 0 THEN
                y -= 1
                LogLine("ERROR: Could not move " + Maildir + "new/" + Maillist(x) + " to " + Maildir + "cur/" + Maillist(x) + "!")
              END IF
            ELSE
              z = KILL(Maildir + "new/" + Maillist(x))
              IF z <> 0 THEN
                y -= 1
                LogLine("ERROR: Could not remove " + Maildir + Maillist(x) + "!")
              END IF
          END IF
        END IF
      NEXT x
      IF UnlockAccount() <> 0 THEN  ' Remove the lock from that account
        LogLine("ERROR: Unlocking the account " + Username + " failed!")
      END IF
      IF y > 0 THEN
          LogLine("" & y & " message(s) deleted.")
          PRINT "+OK " & y & " message(s) deleted in this session. See you later!" + CRLF;
        ELSE
          PRINT "+OK See you later!" + CRLF;
      END IF
      CLOSE #1
      END
    CASE "NOOP"
      PRINT "+OK Yup, still here." + CRLF;
    CASE "CAPA"
      RunCAPA()
    CASE "LAST"
      PRINT "-ERR The LAST command is obsolete. Go get a modern POP3 client!" + CRLF;
    CASE ELSE
      PRINT "-ERR Unrecognised command!" + CRLF;
  END SELECT
LOOP
